#!/usr/bin/env python
# coding: utf-8

from nopea.fields import DbField, ForeignKey, PkField
from nopea.managers import Manager
from nopea.queryset import QuerySet


class MetaType(type):

    def __init__(cls, name, bases, attrs):
        if name == 'DbObject':
            setattr(cls, 'managed_models', [])
            return
        cls.fields = [v for v in cls.__dict__.values() if isinstance(v, DbField)]
        cls.fieldnames = [k for k, v in cls.__dict__.items() if isinstance(v, DbField)]

        pk = [f for f in cls.fields if type(f) == PkField]
        try:
            pk[0]
        except IndexError:
            cls.id = PkField()
            cls.id.value = None
            cls.fields.append(cls.id)
            cls.fieldnames.append('id')


        cls.managed_models.append(cls)

        if hasattr(cls, 'manager'):
            manager = getattr(cls, 'manager')
        else:
            manager = Manager

        cls.objects = manager(cls())

        tablename = getattr(cls, 'tablename', None)
        if not tablename or tablename == 'dbobject':
            cls.tablename = cls.__name__.lower()


class DbObject(metaclass=MetaType):
    adaptor = None
    related_managers = {}

    def __init__(self, *args, **kwargs):
        for name in self.fieldnames:
            field = getattr(self, name)
            if kwargs:
                setattr(self, name, kwargs.get(name))
            setattr(field, 'fieldname', name)
            setattr(field, 'adaptor', self.adaptor)
            if isinstance(field, ForeignKey):
                field.set_related_manager(self)

    def __repr__(self):
        return "<{0}.{1} object>".format(self.__class__.__module__, self.__class__.__name__)

    @classmethod
    def raw(self, query, *args):
        return self.adaptor.execute_query(query, args)[0]

    def get_settings(self):
        settings = {}
        for fieldname in self.fieldnames:
            value = getattr(self, fieldname)
            if isinstance(value, DbObject):
                value = value.id
            settings[fieldname] = value
        return settings

    @classmethod
    def create_table(cls, tablename=None):
        create_table_query = cls.adaptor.get_create_table_query(cls(), tablename)
        return cls.adaptor.execute_query(create_table_query[0], create_table_query[1])

    @classmethod
    def drop_table(cls, tablename=None):
        drop_table_query = cls.adaptor.get_drop_table_query(cls(), tablename)
        return cls.adaptor.execute_query(drop_table_query[0], drop_table_query[1])

    @classmethod
    def add_field(cls, field):
        cls.adaptor.execute_query(field.add_column_query % cls.tablename, None)

    @classmethod
    def drop_field(cls, field, base):
        cls.adaptor.execute_query(field.get_drop_column_query(base), None)

    def delete(self):
        assert self.id is not None, "%s cannot be deleted because it has no id" % self.__class__.__name__
        query_set = QuerySet(self)
        query_set.filter(id=self.id).delete()
        query_set()
        return True

    def save(self) -> bool:
        if self.id:
            query_set = QuerySet(self)
            query_set.filter(id=self.id)
            query_set.partials['updates'] = [self.get_settings()]
            query_set()
            return

        query, query_args = self.adaptor.get_insert_query(self, self.get_settings())
        last_id = self.adaptor.execute_query(query, query_args)[1]
        inst = self.objects.get(id=last_id)
        for name in self.fieldnames:
            setattr(self, name, getattr(inst, name))
        return

    def to_dict(self):
        result = {}
        for name in self.fieldnames:
            field = getattr(self, name)
            if isinstance(field, DbObject):
                field = field.to_dict()
            result[name] = field
        return result



#!/usr/bin/env python
# coding: utf-8

import sqlite3

from nopea.dbobject import DbObject
from nopea.fields import ForeignKey


class IsNull:
    """ Used as identifier for NULL values in query_args """

    def __init__(self):
        pass


class SQLiteAdaptor(object):

    PLACEHOLDER = '?'

    def __init__(self, database):
        self.database = database

    def get_connection_and_cursor(self):
        connection = sqlite3.connect(self.database)
        cursor = connection.cursor()
        return connection, cursor

    def execute_query(self, query, query_args):
        connection, cursor = self.get_connection_and_cursor()
        if not query_args:
            cursor.execute(query)
        else:
            query_args = [q for q in query_args if not isinstance(q, IsNull)]
            cursor.execute(query, query_args)
        connection.commit()
        result = cursor.fetchall()

        cursor.execute("""SELECT last_insert_rowid()""")
        last_id = cursor.fetchone()[0]

        connection.close()
        return (result, last_id)

    def get_filter_query(self, filter_partial):
        key, value = list(filter_partial.items())[0]
        if key.endswith('__in'):
            query = ('%s IN (%s)' % (key.replace('__in', ''), ', '.join('?' * len(value))))
        elif key.endswith('__lt'):
            query = '%s < %s' % (key.replace('__lt', ''), '?')
        elif key.endswith('__lte'):
            query = '%s <= %s' % (key.replace('__lte', ''), '?')
        elif key.endswith('__gt'):
            query = '%s > %s' % (key.replace('__gt', ''), '?')
        elif key.endswith('__gte'):
            query = '%s >= %s' % (key.replace('__gte', ''), '?')
        elif key.endswith('__contains'):
            query = '%s LIKE %s' % (key.replace('__contains', ''), '?')
            value = '%' + value + '%'
        elif key.endswith('__startswith'):
            query = '%s LIKE %s || "%%"' % (key.replace('__startswith', ''), '?')
        elif key.endswith('__endswith'):
            query = '%s LIKE "%%" || %s' % (key.replace('__endswith', ''), '?')
        else:
            if value is not None:
                query = '%s=?' % key
            else:
                query = '%s IS NULL' % key
                value = IsNull()
        return query, value

    def get_exclude_query(self, exclude_partial):
        key, value = list(exclude_partial.items())[0]
        if key.endswith('__in'):
            query = ('%s NOT IN (%s)' % (key.replace('__in', ''), ', '.join('?' * len(value))))
        elif key.endswith('__lt'):
            query = '%s >= %s' % (key.replace('__lt', ''), '?')
        elif key.endswith('__lte'):
            query = '%s > %s' % (key.replace('__lte', ''), '?')
        elif key.endswith('__gt'):
            query = '%s <= %s' % (key.replace('__gt', ''), '?')
        elif key.endswith('__gte'):
            query = '%s < %s' % (key.replace('__gte', ''), '?')
        elif key.endswith('__startswith'):
            query = '%s NOT LIKE %s || "%%"' % (key.replace('__startswith', ''), '?')
        elif key.endswith('__endswith'):
            query = '%s NOT LIKE "%%" || %s' % (key.replace('__endswith', ''), '?')
        else:
            if value is not None:
                query = '%s <> ?' % key
            else:
                value = IsNull()
                query = '%s IS NOT NULL' % key
        return query, value

    def get_limit_query(self, limit):
        return f" LIMIT  {limit}"

    def get_select_query(self, base, *args, **kwargs):
        query = "SELECT %s FROM %s" % (', '.join(base.fieldnames), base.tablename)
        return query

    def get_count_query(self, base, *args, **kwargs):
        query = f'SELECT COUNT(*) FROM "{base.tablename}"'
        return query

    def get_insert_query(self, base, create_partial):
        values = []
        for field in base.fields:
            value = create_partial.get(field.fieldname)
            if isinstance(value, DbObject):
                value = value.id
            if value is None:
                try:
                    value = field.default
                except AttributeError:
                    pass
            values.append(value)
        query = "INSERT INTO %s (%s) VALUES (%s)" % (base.tablename,
                                                     ', '.join(base.fieldnames),
                                                     ', '.join(['?' for item in base.fields]))
        return (query, tuple(values))

    def get_delete_query(self, base):
        query = "DELETE FROM %s " % base.tablename
        return query

    def get_update_query(self, base, update_partial):
        values = []
        affected_fields = []
        settings = ''
        for field in base.fields:
            value = update_partial.get(field.fieldname)
            if not value:
                pass
            else:
                values.append(value)
                affected_fields.append(field.fieldname)

        settings = ', '.join(['%s=?' % item for item in affected_fields])
        query = "UPDATE %s SET %s " % (base.tablename, settings)
        return (query, tuple(values))

    def bulk_create(self, objects, base):
        max_params = 999
        max_inserts_per_query = int(max_params / len(base.fieldnames))
        iterations = (len(base.fieldnames) * len(objects)) / max_params

        lower = 0
        for i in range(0, int(iterations) + 1):
            lower = int(i * max_inserts_per_query)
            upper = int(lower + max_inserts_per_query)
            query, query_args = self.get_bulk_insert_query(objects[lower:upper], base)
            self.execute_query(query, query_args)

    def get_bulk_insert_query(self, objects, base):
        values = [tuple(getattr(item, fieldname) for fieldname in item.fieldnames) for item in objects]
        placeholders = ', '.join(['(%s)' % (', '.join("?" * len(item))) for item in values])
        query = '''INSERT INTO "%s" (%s) VALUES %s''' % (base.tablename, ', '.join(base.fieldnames), placeholders)
        return (query, tuple(v for tupl in values for v in tupl))

    def get_create_table_query(self, base, tablename):
        if not tablename:
            tablename = base.tablename
        query = "CREATE TABLE %s (%%s);" % tablename
        fk_fields = []
        fields = []
        fieldnames = []
        for field in base.fields:
            if isinstance(field, ForeignKey):
                fk_fields.append(field)
            fields.append(field)
            fieldnames.append(field.fieldname)

        field_queries = [field.partial_create_table_query for field in fields]

        for field in fk_fields:
            fieldnames.append(field.fieldname)
        field_queries.extend([field.partial_create_table_query_extension for field in fk_fields])

        query = query % ', '.join(field_queries)
        query = query % tuple(fieldnames)
        return (query, None)

    def get_drop_table_query(self, base, tablename):
        if not tablename:
            tablename = base.tablename
        query = "DROP TABLE %s;" % tablename
        return (query, None)

    def get_drop_column_query(self, field, base):
        tablename = base.tablename
        fieldnames = base.fieldnames
        fieldnames.remove(field.fieldname)
        delattr(base, field.fieldname)
        queries_1 = ["CREATE TABLE t_backup({0})".format(', '.join(fieldnames)),
                     "INSERT INTO t_backup SELECT {0} FROM {1}".format(', '.join(fieldnames), tablename),
                     "DROP TABLE {0}".format(tablename)]

        queries_2 = ["INSERT INTO {1} SELECT {0} FROM t_backup".format(', '.join(fieldnames), tablename),
                     "DROP TABLE t_backup",
                     "VACUUM"]
        for query in queries_1:
            self.execute_query(query, None)

        base.create_table()

        for query in queries_2:
            self.execute_query(query, None)
        return "SELECT NULL"

    def get_pkfield_create_query(self):
        return '%s INTEGER PRIMARY KEY AUTOINCREMENT'

    def get_integer_field_create_query(self, field):
        query = '%s INTEGER'
        if field.default is not None:
            query += ' DEFAULT %s' % field.default
        return query

    def get_integer_field_create_column_query(self, field):
        query = 'ALTER TABLE %%s ADD COLUMN %s INTEGER' % field.fieldname
        if field.default is not None:
            query += ' DEFAULT %s' % field.default
        return query

    def get_char_field_create_query(self, field):
        query = '%%s CHAR(%s)' % field.max_length
        if field.default is not None:
            query += ' DEFAULT %s' % field.default
        return query

    def get_charfield_create_column_query(self, field):
        query = 'ALTER TABLE %%s ADD COLUMN %s CHAR(%s)' % (field.fieldname, field.max_length)
        if field.default is not None:
            query += ' DEFAULT %s' % field.default
        return query

    def get_text_field_create_query(self):
        return '%s TEXT'

    def get_text_field_create_column_query(self, field):
        return 'ALTER TABLE %%s ADD COLUMN %s TEXT' % (field.fieldname)

    def get_datetime_field_create_query(self, default):
        if default is None:
            default = 'NULL'
        elif default == 'now':
            default = 'CURRENT_TIMESTAMP'
        return '%%s TIMESTAMP DEFAULT %s' % default

    def get_boolean_field_create_query(self, default):
        return '%%s TINYINT DEFAULT %s' % default

    def get_boolean_field_create_column_query(self, field):
        query = 'ALTER TABLE %%s ADD COLUMN %s TINYINT DEFAULT %s' % (field.fieldname, field.default)
        return query

    def get_foreignkey_field_create_query_base(self):
        return '%s INTEGER'

    def get_foreignkey_field_create_query_extension(self, reference):
        if type(DbObject) == type(reference):
            reference = reference.tablename
        return 'FOREIGN KEY(%%s) REFERENCES %s(id)' % (reference)

    def create_migration_table(self):
        query = """CREATE TABLE IF NOT EXISTS nopea_migrations
                    (name CHAR(250), id INTEGER PRIMARY KEY AUTOINCREMENT)"""
        self.execute_query(query, None)

    def get_done_migrations(self):
        query = "SELECT id, name FROM nopea_migrations"
        return self.execute_query(query, None)

    def insert_migration(self, name):
        query = "INSERT INTO nopea_migrations (name) VALUES (?)"
        return self.execute_query(query, (name,))



# This file contains example model definitions.
from nopea import fields
from app import DbObject


class User(DbObject):

    def __repr__(self):
        return "<{0}: {1}>".format(self.__class__.__name__, self.username)

    username = fields.CharField(max_length=250)
    first_name = fields.CharField(max_length=250)


class Address(DbObject):
    a = fields.IntegerField()



#!/usr/bin/env python
# coding: utf-8

import importlib
import os
import sys

from collections import defaultdict
from pprint import pformat

from datetime import datetime

from nopea.dbobject import DbObject
from nopea.fields import BooleanField, CharField, DateTimeField, ForeignKey, IntegerField, PkField, TextField

field_types = [BooleanField, CharField, DateTimeField, ForeignKey, IntegerField, PkField, TextField]
field_dict = {field.__name__: field for field in field_types}


class Migration:
    migration_dir = None

    def __init__(self):
        self.tablename = 'nopea_migrations'
        self.adaptor = DbObject.adaptor
        self.managed_models = DbObject.managed_models
        self.ref_models = {model.__name__: model for model in self.managed_models}
        self.current_model_state = {}
        self.adaptor.create_migration_table()
        self.render_current_model_state()
        sys.path.append(self.migration_dir)

    def render_current_model_state(self):
        for model in self.managed_models:
            instance = model()
            class_name = instance.__class__.__name__
            self.current_model_state[class_name] = {'tablename': model.tablename, 'fields': []}
            for field in instance.fields:
                field_attrs = {}
                if getattr(field, 'default', None) is not None:
                    field_attrs['default'] = getattr(field, 'default', None)
                if getattr(field, 'max_length', None) is not None:
                    field_attrs['max_length'] = getattr(field, 'max_length', None)
                if getattr(field, 'reference_class', None) is not None:
                    ref_obj = getattr(field, 'reference_class', None)
                    field_attrs['reference_class'] = ref_obj.__name__
                self.current_model_state[class_name]['fields'].append(
                    {field.fieldname: {
                        'field_type': field.__class__.__name__,
                        'field_attrs': field_attrs
                    }}
                )

        return self.current_model_state

    def get_field_changes(self, old_state, new_state):
        fields_to_add = defaultdict(list)
        fields_to_remove = defaultdict(list)

        models_to_check = new_state.keys() & old_state.keys()
        for model in models_to_check:
            old_field_dict = {}
            old_state_field_names = set()
            for field in old_state[model]['fields']:
                for key, value in field.items():
                    old_state_field_names.add(key)
                    old_field_dict[key] = value

            new_field_dict = {}
            new_state_field_names = set()
            for field in new_state[model]['fields']:
                for key, value in field.items():
                    new_state_field_names.add(key)
                    new_field_dict[key] = value

            new_fields = new_state_field_names - old_state_field_names
            dispensable_fields = old_state_field_names - new_state_field_names

            for field in new_fields:
                fields_to_add[model].append({field: new_field_dict[field]})

            for field in dispensable_fields:
                fields_to_remove[model].append({field: old_field_dict[field]})
        return (dict(fields_to_add), dict(fields_to_remove))

    def render_actions(self, old_state, new_state):
        actions = {}
        models_to_create = {model: new_state[model] for model in (new_state.keys() - old_state.keys())}
        if models_to_create:
            actions['creations'] = models_to_create

        models_to_delete = {model: old_state[model] for model in (old_state.keys() - new_state.keys())}
        if models_to_delete:
            actions['deletions'] = models_to_delete

        fields_to_add, fields_to_remove = self.get_field_changes(old_state, new_state)

        if fields_to_add:
            actions['fields_to_add'] = fields_to_add
        if fields_to_remove:
            actions['fields_to_remove'] = fields_to_remove
        return actions

    def get_existing_migration_files(self):
        existing_migration_files = []
        for f in os.listdir(self.migration_dir):
            if f.endswith('.py') and not f == '__init__.py':
                existing_migration_files.append(f)
        existing_migration_files.sort()
        return existing_migration_files

    def get_last_migration(self, existing_migration_files):
        if existing_migration_files:
            return existing_migration_files[-1]

    def get_content_from_migration_file(self, migration_file):
        old_state = {}
        actions = {}
        callables = []
        new_state = {}
        if migration_file:
            content = importlib.import_module(migration_file.split('.')[0])
            new_state = content.new_state  # MUST be there

            try:
                old_state = content.old_state
            except AttributeError:
                pass
            try:
                actions = content.actions
            except AttributeError:
                pass
                new_state = content.new_state
            try:
                callables = content.callables
            except AttributeError:
                pass

        return (old_state, actions, new_state, callables)

    def create_migrations(self):
        timestamp = datetime.strftime(datetime.now(), '%Y%m%d_%H%M%S')
        existing_migration_files = self.get_existing_migration_files()
        last_migration_number = len(existing_migration_files)
        last_migration = self.get_last_migration(existing_migration_files)
        _, _, new_state, _ = self.get_content_from_migration_file(last_migration)
        old_state = new_state

        migration_file_path = os.path.join(
            self.migration_dir, str(last_migration_number + 1).zfill(3) + '_' + timestamp + '.py'
        )

        actions = self.render_actions(old_state, self.current_model_state)
        if not actions:
            print("No changes detected")
            return

        if actions.get('creations'):
            print("Creations: {0}".format(', '.join(actions.get('creations').keys())))

        if actions.get('deletions'):
            print("Deletions: {0}".format(', '.join(actions.get('deletions').keys())))

        if actions.get('fields_to_add'):
            print("Fields to add: {0}".format(', '.join(actions.get('fields_to_add').keys())))

        if actions.get('fields_to_remove'):
            print("Fields to remove: {0}".format(', '.join(actions.get('fields_to_remove').keys())))

        with open(migration_file_path, 'w') as migration_file:
            old_state_string = "old_state = " + pformat(old_state) + "\n\n"
            actions_string = "actions = " + pformat(actions) + "\n\n"
            new_state_string = "new_state = " + pformat(self.current_model_state) + "\n\n"
            content = old_state_string + actions_string + new_state_string
            migration_file.write(content)

    def get_dummy(self, attrs):
        dummy = DbObject
        setattr(dummy, 'tablename', attrs['tablename'])
        setattr(dummy, 'fieldnames', [])
        for field in attrs['fields']:
            for fieldname, field_conf in field.items():
                dummy.fieldnames.append(fieldname)
                field_attrs = field_conf['field_attrs']
                if field_attrs.get('reference_class'):
                    ref_cls = field_attrs.get('reference_class')
                    field_attrs['reference_class'] = self.ref_models[ref_cls]
                field_cls = field_dict.get(field_conf.get('field_type'))
                field_inst = field_cls(**field_attrs)
                setattr(field_inst, 'adaptor', self.adaptor)
                setattr(field_inst, 'fieldname', fieldname)
                setattr(dummy, fieldname, field_inst)
        return dummy

    def create_table(self, attrs):
        migration_dummy = self.get_dummy(attrs)
        migration_dummy.create_table()

    def drop_table(self, attrs):
        migration_dummy = self.get_dummy(attrs)
        migration_dummy.drop_table()

    def add_fields(self, model, fields, old_state):
        attrs = old_state[model]
        migration_dummy = self.get_dummy(attrs)
        for field in fields:
            field_type = field_dict[list(field.values())[0]['field_type']]
            field_attrs = list(field.values())[0]['field_attrs']
            field_dummy = field_type(**field_attrs)
            field_dummy.fieldname = list(field.keys())[0]
            field_dummy.adaptor = self.adaptor
            migration_dummy.add_field(field_dummy)

    def drop_fields(self, model, fields, old_state):
        attrs = old_state[model]
        migration_dummy = self.get_dummy(attrs)
        for field in fields:
            field_type = field_dict[list(field.values())[0]['field_type']]
            field_attrs = list(field.values())[0]['field_attrs']
            field_dummy = field_type(**field_attrs)
            field_dummy.fieldname = list(field.keys())[0]
            field_dummy.adaptor = self.adaptor
            migration_dummy.drop_field(field_dummy, migration_dummy)
        pass

    def run_migrations(self):
        existing_migration_files = self.get_existing_migration_files()
        done_migrations = [name[1] for name in self.adaptor.get_done_migrations()[0]]
        migrations_to_run = [f for f in existing_migration_files if f not in done_migrations]
        if not migrations_to_run:
            print("Nothing to migrate")
        for migration_file in migrations_to_run:
            print("Running migration: %s" % migration_file)
            old_state, actions, new_state, callables = self.get_content_from_migration_file(migration_file)

            for creation in actions.get('creations', {}).items():
                self.create_table(creation[1])
            for deletion in actions.get('deletions', {}).items():
                self.drop_table(deletion[1])

            for field_creation in actions.get('fields_to_add', {}).items():
                self.add_fields(field_creation[0], field_creation[1], old_state)

            for field_deleteion in actions.get('fields_to_remove', {}).items():
                self.drop_fields(field_deleteion[0], field_deleteion[1], old_state)

            for func in callables:
                func()

            self.adaptor.insert_migration(migration_file)
